@@ -3,7 +3,7 @@
 # Copyright © 2008-2013  Red Hat, Inc. All rights reserved.
 # Copyright © 2008-2013  Luke Macken <lmacken@redhat.com>
 # Copyright © 2008  Kushal Das <kushal@fedoraproject.org>
-# Copyright © 2012-2015  Tails Developers <tails@boum.org>
+# Copyright © 2012-2016  Tails Developers <tails@boum.org>
 #
 # This copyrighted material is made available to anyone wishing to use, modify,
 # copy, or redistribute it subject to the terms and conditions of the GNU
@@ -193,6 +193,13 @@ class TailsInstallerThread(threading.Thread):
                 if not self.live.can_read_partition_table():
                     self.live.log.info('Clearing unreadable partition table.')
                     self.live.clear_all_partition_tables()
+                if self.parent.force_reinstall:
+                    parent = _to_unicode(self.live.drive['parent'])
+                    self.rescan_devices(force_partitions=False)
+                    if parent in self.live.drives and \
+                       'device' in self.live.drives[parent]:
+                        self.live.drive = self.live.drives[parent]['device']
+                        self.live.save_full_drive()
                 self.live.partition_device()
                 self.rescan_devices(force_partitions=True)
                 self.live.switch_drive_to_system_partition()
@@ -227,8 +234,6 @@ class TailsInstallerThread(threading.Thread):
             self.live.create_persistent_overlay()
             self.live.update_configs()
 
-            if not self.parent.opts.partition and self.live.is_partition_GPT():
-                self.live.update_system_partition_properties()
             self.live.install_bootloader()
             # self.live.bootable_partition()
 
@@ -254,7 +259,7 @@ class TailsInstallerThread(threading.Thread):
             self.installation_complete()
 
         except Exception, e:
-            self.status(e.args[0])
+            self.status(e)
             self.status(_("Tails installation failed!"))
             self.live.log.exception(unicode(e))
             self.live.log.debug(traceback.format_exc())
@@ -296,52 +301,53 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         self.source_available = False
         self.target_available = False
         self.target_selected  = False
-        self.persistence = False
+        self.devices_with_persistence = []
+        self.force_reinstall = False
+
+        if self.opts.unprivileged == None:
+            self.opts.unprivileged = True
+        if self.opts.noverify == None:
+            self.opts.noverify = True
 
         self._build_ui()
 
-        self.update_start_button()
-        if self.opts.clone or config['download']['enabled']:
-            self.source_available = True
-        if self.opts.clone:
-            self.__box_source.set_visible(False)
-        if not config['download']['enabled']:
-            self.__box_source_dl.set_visible(False)
-        # FIXME: this may be useful when we'll support windows
-        #if sys.platform == 'win32':
-        #    self.driveBox.setGeometry(QtCore.QRect(10, 20, 145, 22))
-        #    self.refreshDevicesButton = QtGui.QPushButton(self.targetGroupBox)
-        #    self.refreshDevicesButton.setGeometry(QtCore.QRect(156, 16, 30, 26))
-        #    self.refreshDevicesButton.setText("")
-        #    icon = QtGui.QIcon()
-        #    icon.addPixmap(QtGui.QPixmap(":/refresh.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
-        #    self.refreshDevicesButton.setIcon(icon)
-        #    self.refreshDevicesButton.setFlat(True)
-        #    self.refreshDevicesButton.setObjectName("refreshDevicesButton")
-        self.update_start_button()
+        self.opts.clone = is_running_from_tails()
         self.live = TailsInstallerCreator(opts=opts)
-        if self.opts.clone:
-            self.live.source = RunningLiveSystemSource(
-                path=config['running_liveos_mountpoint'])
-        if not self.opts.clone:
-            self.populate_releases()
-        self.populate_devices()
-        self.downloader = None
-        self.progress_thread = ProgressThread(parent=self)
-        self.download_progress = DownloadProgress(parent=self)
-        self.live_thread = TailsInstallerThread(live=self.live,
-                                         progress=self.progress_thread,
-                                         parent=self)
-        self.live.connect_drive_monitor(self.populate_devices)
-        self.confirmed = False
-        self.delete_existing_liveos_confirmed = False
 
-        # Intercept all tails_installer.INFO log messages, and display them in the gui
+        # Intercept all tails_installer.INFO log messages, and display them
+        # in the GUI
         self.handler = TailsInstallerLogHandler(lambda x: self.append_to_log(x))
         self.live.log.addHandler(self.handler)
         if not self.opts.verbose:
             self.live.log.removeHandler(self.live.stream_handler)
 
+        # Initialize the visibility and state of the clone/ISO source
+        # selection widgets. This triggers on_radio_button_source_iso_toggled
+        # that will in turn initialize self.live.source, self.source_available,
+        # and populate the list of candidate devices.
+        # - inside Tails
+        if self.opts.clone:
+            self.__radio_button_source_device.set_active(True)
+            self.__filechooserbutton_source_file.set_sensitive(False)
+        # - outside of Tails
+        else:
+            self.__radio_button_source_device.set_visible(False)
+            self.__filechooserbutton_source_file.set_sensitive(True)
+
+        self.__box_source_dl.set_visible(False)
+        if not self.opts.clone:
+            self.populate_releases()
+        self.update_start_button()
+        self.downloader = None
+        self.progress_thread = ProgressThread(parent=self)
+        self.download_progress = DownloadProgress(parent=self)
+        self.live_thread = TailsInstallerThread(live=self.live,
+                                                progress=self.progress_thread,
+                                                parent=self)
+        self.live.connect_drive_monitor(self.populate_devices)
+        self.confirmed = False
+        self.delete_existing_liveos_confirmed = False
+
         # If an ISO was specified on the command line, use it.
         if args:
             for arg in self.args:
@@ -373,8 +379,9 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         self.__label_infobar_title = builder.get_object('label_infobar_title')
         self.__label_infobar_details = builder.get_object('label_infobar_details')
         self.__box_source = builder.get_object('box_source')
+        self.__box_source_select = builder.get_object('box_source_select')
         self.__box_source_file = builder.get_object('box_source_file')
-        self.__filechooserbutton_source_file = builder.get_object('liststore_source')
+        self.__filechooserbutton_source_file = builder.get_object('filechooserbutton_source_file')
         self.__box_source_dl = builder.get_object('box_source_dl')
         self.__liststore_source_dl = builder.get_object('liststore_source_dl')
         self.__combobox_source_dl = builder.get_object('combobox_source_dl')
@@ -384,6 +391,9 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         self.__textview_log = builder.get_object('textview_log')
         self.__progressbar = builder.get_object('progressbar_progress')
         self.__button_start = builder.get_object('button_start')
+        self.__radio_button_source_iso = builder.get_object('radio_button_source_iso')
+        self.__radio_button_source_device = builder.get_object('radio_button_source_device')
+        self.__button_force_reinstall = builder.get_object('check_force_reinstall')
 
         self.add(self.__box_installer)
         builder.connect_signals(self)
@@ -404,10 +414,90 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         rgba.parse(config['branding']['color'])
         self.__image_header.override_background_color(Gtk.StateFlags.NORMAL, rgba)
 
+    def on_radio_button_source_iso_toggled(self, radio_button):
+        self.live.log.debug("Entering on_radio_button_source_iso_toggled")
+        active_radio = [r for r in radio_button.get_group() if r.get_active()][0]
+        if active_radio.get_label() == _("Clone the current Tails"):
+            self.live.log.debug("Mode: clone")
+            self.opts.clone = True
+            self.live.source = RunningLiveSystemSource(
+                path=config['running_liveos_mountpoint'])
+            self.source_available = True
+            self.__filechooserbutton_source_file.set_sensitive(False)
+        elif active_radio.get_label() == _("Use a downloaded Tails ISO image"):
+            self.live.log.debug("Mode: from ISO")
+            self.opts.clone = False
+            self.live.source = None
+            self.source_available = False
+            self.__filechooserbutton_source_file.set_sensitive(True)
+        # previous error messages may be invalid now
+        self.clear_log()
+        # some previously rejected devices may now be valid candidates
+        # and vice-versa
+        self.live.log.debug("Calling populate_devices()"
+                            " from on_radio_button_source_iso_toggled")
+        self.populate_devices()
+
+    def on_force_reinstall_clicked(self, button):
+        # If the user has chosen install from ISO, but no ISO is selected
+        if not self.live.opts.clone and not self.is_ISO_selected():
+            self.warn_ISO_not_selected()
+            return
+        self.force_reinstall = True
+        self.opts.partition = True
+        self.live.force_reinstall = True
+        self.__button_force_reinstall.set_visible(False)
+        self.on_start_clicked(button)
+
     def on_source_file_set(self, filechooserbutton):
         self.select_source_iso(filechooserbutton.get_filename())
 
+    def on_target_changed(self, combobox_target):
+        # get selected device
+        self.__button_force_reinstall.set_visible(False)
+        drive = self.get_selected_drive()
+        if drive == None:
+           return
+        device = self.live.drives[drive]
+
+        if self.live.device_can_be_upgraded(device):
+            self.opts.partition = False
+            self.force_reinstall = False
+            self.live.force_reinstall = False
+            self.__button_start.set_label(_('Upgrade'))
+            self.__button_force_reinstall.set_visible(True)
+        else:
+            self.opts.partition = True
+            self.force_reinstall = True
+            self.live.force_reinstall = False
+            self.__button_start.set_label(_('Install'))
+            self.__button_force_reinstall.set_visible(False)
+
+    def get_device_pretty_name(self, device):
+        size = _format_bytes_in_gb(device['parent_size']
+                                    if device['parent_size']
+                                    else device['size'])
+        pretty_name = _("%(size)s %(vendor)s %(model)s device (%(device)s)") % {
+                    'size':   size,
+                    'vendor': device['vendor'],
+                    'model':  device['model'],
+                    'device': device['device']
+        }
+        return pretty_name
+
+    def is_ISO_selected(self):
+        return (self.live.source.__class__ == LocalIsoSource)
+
+    def warn_ISO_not_selected(self):
+        self.show_confirmation_dialog(_("No ISO image selected"),
+                                      _("Please select a Tails ISO image."),
+                                        True)
+
     def on_start_clicked(self, button):
+        # If the user has chosen install from ISO, but no ISO is selected
+        if not self.live.opts.clone and not self.is_ISO_selected():
+            self.warn_ISO_not_selected()
+            return
         self.begin()
 
     def on_infobar_response(self, infobar, response):
@@ -415,16 +505,23 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         self.__label_infobar_title.set_text("")
         self.__label_infobar_details.set_text("")
 
+    def clear_log(self):
+        text_buffer = self.__textview_log.get_buffer()
+        text_buffer.set_text('')
+
     def append_to_log(self, text):
         if not text.endswith('\n'):
             text = text + '\n'
         text_buffer = self.__textview_log.get_buffer()
         text_buffer.insert(text_buffer.get_end_iter(), text)
-        self.__textview_log.scroll_to_iter(text_buffer.get_end_iter(), 0, False, 0, 0)
+        self.__textview_log.scroll_to_iter(text_buffer.get_end_iter(),
+                                           0, False, 0, 0)
 
     def update_start_button(self):
         if self.source_available and self.target_available:
             self.__button_start.set_sensitive(True)
+            if not self.opts.partition:
+                self.__button_force_reinstall.set_visible(True)
         else:
             self.__button_start.set_sensitive(False)
 
@@ -440,8 +537,7 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
                         _("No device suitable to install Tails could be found"))
                 self.__label_infobar_details.set_text(
                         _("Please plug a USB flash drive or SD card of at least %0.1f GB.")
-                        % ((config['min_system_partition_size'] +
-                            config['min_persistence_partition_size']) / 1000.))
+                        % (config['official_min_installation_device_size'] / 1000.))
                 self.__infobar.set_visible(True)
                 self.target_available = False
                 self.update_start_button()
@@ -449,6 +545,8 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
             else:
                 self.__infobar.set_visible(False)
             self.live.log.debug('drives: %s' % self.live.drives)
+            target_list = []
+            self.devices_with_persistence = []
             for device, info in self.live.drives.items():
                 # Skip the device that is the source of the copy
                 if (
@@ -467,22 +565,9 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
                 # Skip LUKS-encrypted partitions
                 if info['fstype'] and info['fstype'] == 'crypto_LUKS':
                     self.live.log.debug('Skipping LUKS-encrypted partition: %s' % info['device'])
+                    self.devices_with_persistence.append(info['parent'])
                     continue
-                size = _format_bytes_in_gb(info['parent_size']
-                                           if info['parent_size']
-                                           else info['size'])
-                details = (_("%(size)s %(label)s") % {
-                                'label': info['label'],
-                                'size': size
-                            }
-                           if info['label']
-                           else size)
-                pretty_name = _("%(vendor)s %(model)s (%(details)s) - %(device)s") % {
-                    'device':  info['device'],
-                    'vendor':  info['vendor'],
-                    'model':   info['model'],
-                    'details': details
-                }
+                pretty_name = self.get_device_pretty_name(info)
                 # Skip devices with non-removable bit enabled
                 if not info['removable']:
                     message =_('The USB stick "%(pretty_name)s"'
@@ -494,33 +579,32 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
                     self.status(message)
                     continue
                 # Skip too small devices, but inform the user
-                if not info['is_device_big_enough']:
+                if not info['is_device_big_enough_for_installation']:
                     message =_('The device "%(pretty_name)s"'
                                ' is too small to install'
-                               ' Tails (at least %(size)s GB is required).') %  {
-                               'pretty_name':  pretty_name,
-                               'size'  :  (float(config['min_system_partition_size']
-                                       + config['min_persistence_partition_size'])
-                                       / 1000)
+                               ' Tails (at least %(size)s GB is required).') % {
+                               'pretty_name': pretty_name,
+                               'size': (float(
+                                   config['official_min_installation_device_size'])
+                                        / 1000)
                                }
                     self.status(message)
                     continue
-                # Skip devices without a Tails installation
-                if not self.opts.partition and not self.live.device_can_be_upgraded(info):
-                    if is_running_from_tails():
-                        action = _('\"Install by cloning\"')
-                    else:
-                        action = _('\"Install from ISO\"')
-                    message = _('It is impossible to upgrade the device %(pretty_name)s'
-                               ' because it was not created using Tails Installer.'
-                               ' You should instead use %(action)s to upgrade Tails'
-                               ' on this device.') % {
-                               'pretty_name' : pretty_name,
-                               'action'      : action
-                    }
+                # Skip devices too small for cloning, but inform the user
+                if self.opts.clone \
+                   and not info['is_device_big_enough_for_upgrade']:
+                    message = _('To upgrade device "%(pretty_name)s"'
+                                ' from this Tails, you need to use'
+                                " a downloaded Tails ISO image:\n"
+                                'https://tails.boum.org/install/download') % {
+                                    'pretty_name': pretty_name,
+                                }
                     self.status(message)
                     continue
-                self.__liststore_target.append([pretty_name, device])
+                target_list.append([pretty_name, device])
+            if len(target_list):
+                for target in target_list:
+                    self.__liststore_target.append(target)
                 self.target_available = True
                 self.__combobox_target.set_active(0)
                 self.update_start_button()
@@ -552,18 +636,30 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
     def progress(self, value):
         self.__progressbar.set_fraction(value)
 
-    def status(self, text):
-        if isinstance(text, Exception):
-            text = text.args[0]
-        elif isinstance(text, int):
-            text = str(text)
-        self.append_to_log(text)
+    def status(self, obj):
+        try:
+            if isinstance(obj, str) or isinstance(obj, unicode):
+                text = obj
+            elif isinstance(obj, Exception) \
+                 and hasattr(obj, 'args') \
+                 and type(obj.args).__name__ == 'list':
+                text = obj.args[0]
+            else:
+                text = str(obj)
+            self.append_to_log(text)
+        except Exception as e:
+            self.live.log.exception(
+                "Failed to set status to object of type '{type}'"
+                .format(type=type(obj).__name__)
+            )
+            raise e
 
     def enable_widgets(self, enabled=True):
         if enabled:
             self.update_start_button()
         else:
             self.__button_start.set_sensitive(False)
+            self.__button_force_reinstall.set_visible(False)
         self.__box_source.set_sensitive(enabled)
         self.__combobox_source_dl.set_sensitive(enabled)
         self.__combobox_target.set_sensitive(enabled and not self.target_selected)
@@ -575,8 +671,10 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         self.in_process = not enabled
 
     def get_selected_drive(self):
-        drive = self.__liststore_target.get(
-                self.__combobox_target.get_active_iter(), 1)[0]
+        drive = None
+        iter = self.__combobox_target.get_active_iter()
+        if iter != None:
+            drive = self.__liststore_target.get(iter, 1)[0]
         if drive:
             return _to_unicode(drive)
 
@@ -587,15 +685,18 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
                                    message_type=Gtk.MessageType.INFO,
                                    buttons=Gtk.ButtonsType.CLOSE,
                                    message_format=_("Installation complete!"))
-        dialog.format_secondary_text(_("Installation was completed."))
         dialog.run()
         self.close()
 
-    def show_confirmation_dialog(self, title, message):
+    def show_confirmation_dialog(self, title, message, warning):
+        if (warning):
+            buttons=Gtk.ButtonsType.OK
+        else:
+            buttons=Gtk.ButtonsType.YES_NO
         dialog = Gtk.MessageDialog(parent=self,
                                    flags=Gtk.DialogFlags.DESTROY_WITH_PARENT,
                                    message_type=Gtk.MessageType.QUESTION,
-                                   buttons=Gtk.ButtonsType.YES_NO,
+                                   buttons=buttons,
                                    message_format=title)
         dialog.format_secondary_text(message)
         reply = dialog.run()
@@ -639,16 +740,15 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         if self.opts.partition:
             if not self.confirmed:
                 if self.show_confirmation_dialog(
-                        _("Please confirm your device selection"),
-                        _("You are going to install Tails on the "
-                          "%(size)s %(vendor)s %(model)s device (%(device)s). "
-                          "All data on the selected device will be lost. "
-                          "Continue?") %
+                        _("Confirm the target USB stick"),
+                        _("%(size)s %(vendor)s %(model)s device (%(device)s)\n\n"
+                          "All data on this USB stick will be lost.") %
                           {'vendor': self.live.drive['vendor'],
                            'model':  self.live.drive['model'],
                            'device': self.live.drive['device'],
                            'size':   _format_bytes_in_gb(self.live.drive['size'])
-                          }):
+                          },
+                          False):
                     self.confirmed = True
                 else:
                     return
@@ -657,17 +757,20 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
                 # let's go on
                 self.confirmed = False
         else:
-            msg = (_("You are going to upgrade Tails on the %(parent_size)s "
-                     "%(vendor)s %(model)s device (%(device)s). "
-                     "Any persistent volume on this device will remain unchanged. "
-                     "Continue?") % {
-                         'vendor': self.live.drive['vendor'],
-                         'model':  self.live.drive['model'],
-                         'device': self.live.drive['device'],
-                         'parent_size': _format_bytes_in_gb(self.live.drive['parent_size'])
-                     }
-                  )
-            if self.show_confirmation_dialog(_("Please confirm your device selection"), msg):
+            description = _("%(parent_size)s %(vendor)s %(model)s device (%(device)s)") % {
+                'vendor': self.live.drive['vendor'],
+                'model':  self.live.drive['model'],
+                'device': self.live.drive['device'],
+                'parent_size': _format_bytes_in_gb(self.live.drive['parent_size']),
+            }
+            persistence_message = ""
+            if self.devices_with_persistence:
+                persistence_message = _("\n\nThe persistent storage on this USB stick will be preserved.")
+            msg = _("%(description)s%(persistence_message)s") % {
+                'description': description,
+                'persistence_message': persistence_message,
+            }
+            if self.show_confirmation_dialog(_("Confirm the target USB stick"), msg, False):
                 # The user has confirmed that they wish to overwrite their
                 # existing Live OS.  Here we delete it first, in order to
                 # accurately calculate progress.
@@ -693,14 +796,6 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
         elif self.live.source.__class__ == LocalIsoSource:
             self.enable_widgets(False)
             self.live_thread.start()
-        # Else, download an ISO.
-        elif config['download']['enabled']:
-            self.downloader = ReleaseDownloader(
-                    self.downloadCombo.currentText(),
-                    progress=self.download_progress,
-                    proxies=self.live.get_proxies(),
-                    parent=self)
-            self.downloader.start()
         else:
             raise NotImplementedError
 
@@ -739,9 +834,11 @@ class TailsInstallerWindow(Gtk.ApplicationWindow):
             return False
 
         self.live.log.info(_("%(filename)s selected")
-                             % {'filename': os.path.basename(self.live.source.path)})
+                             % {'filename': str(os.path.basename(self.live.source.path))})
         self.source_available = True
-        self.update_start_button()
+        self.live.log.debug("Calling populate_devices()"
+                            " from select_source_iso")
+        self.populate_devices()
 
     def terminate(self):
         """ Final clean up """
